# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import os
import numpy as np
import copy

from l2rpn_baselines.utils import NNParam
from l2rpn_baselines.DuelQLeapNet.DuelQLeapNet_NN import DuelQLeapNet_NN


class LeapNet_NNParam(NNParam):
    """
    This class implements the type of parameters used by the DuelQLeapNet model.

    More information on the leap net can be found at `Leap Net on Github <https://github.com/BDonnot/leap_net>`_

    Attributes
    -----------
    x_dim: ``int``
        Dimension of the input `x`

    list_attr_obs_tau: ``str``
        List of the name of the observation variable that will be used as vector tau to performs the leaps.

    tau_dims: ``list``
        List of ``int``. For each variable considered as a `tau` specify its dimension here.

    tau_adds: ``list``
        List of ``float`` if you want to add something to the value of the observation you receive. For example if you
        know the observation you will receive is either 1 or 2 but prefer these number to be 0 and 1, you can set
        the relevant `tau_adds` to "-1"

    tau_mults: ``list``
        List of ``float``. Same as above if for multiplicative term. If you want to multiply the number you get by a
        specific number (for example if you have numbers in the range 0,10 but would rather have numbers in the range
        0,1, you can set the `tau_mults` number to `0.1`

    """
    _int_attr = copy.deepcopy(NNParam._int_attr)
    _float_attr = copy.deepcopy(NNParam._float_attr)
    _str_attr = copy.deepcopy(NNParam._str_attr)
    _list_float = copy.deepcopy(NNParam._list_float)
    _list_str = copy.deepcopy(NNParam._list_str)
    _list_int = copy.deepcopy(NNParam._list_int)

    _int_attr += ["x_dim"]
    _list_str += ["list_attr_obs_tau"]
    _list_float += ["tau_adds", "tau_mults"]
    _list_int += ["tau_dims"]
    nn_class = DuelQLeapNet_NN

    def __init__(self,
                 action_size,
                 observation_size,  # not used here for retro compatibility with NNParam.from_dict
                 sizes,
                 activs,
                 x_dim,
                 list_attr_obs,
                 tau_dims,
                 tau_adds,
                 tau_mults,
                 list_attr_obs_tau,
                 ):
        NNParam.__init__(self,
                         action_size,
                         observation_size=x_dim + np.sum(tau_dims),  # TODO this might not be usefull
                         sizes=sizes,
                         activs=activs,
                         list_attr_obs=list_attr_obs
                         )
        self.tau_dims = [int(el) for el in tau_dims]
        self.list_attr_obs_tau = [str(el) for el in list_attr_obs_tau]
        self.x_dim = x_dim
        self.tau_adds = tau_adds
        self.tau_mults = tau_mults

    def get_obs_attr(self):
        return self.list_attr_obs + self.list_attr_obs_tau


"""
__author__ = 'jiajun,yanzan'
# Copyright (C) 2018 - 2020 GEIRI North America
# Authors: jiajaun <jiajun.duan@geirina.net>
"""

from grid2op.Agent import DoNothingAgent


import time
import datetime
import warnings
import copy
import itertools

import sys
import csv

import os
import json
import math
import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

from grid2op.Parameters import Parameters
from grid2op.Agent import AgentWithConverter
from grid2op.Converter import IdToAct

from l2rpn_baselines.Geirina.DDQN_NN import DeepQNetworkDueling
from l2rpn_baselines.Geirina.prioritized_memory import Memory


class Geirina(DoNothingAgent):
    """
    Do nothing agent of grid2op, as a lowerbond baseline for l2rpn competition.

    Note that a Baseline should always somehow inherit from :class:`grid2op.Agent.BaseAgent`.

    It serves as a template agent to explain how a baseline can be built.

    As opposed to bare grid2op Agent, baselines have 3 more methods:
    - :func:`Template.load`: to load the agent, if applicable
    - :func:`Template.save`: to save the agent, if applicable
    - :func:`Template.train`: to train the agent, if applicable

    The method :func:`Template.reset` is already present in grid2op but is emphasized here. It is called
    by a runner at the beginning of each episode with the first observation.

    The method :func:`Template.act` is also present in grid2op, of course. It the main method of the baseline,
    that receives an observation (and a reward and flag that says if an episode is over or not) an return a valid
    action.

    **NB** the "real" instance of environment on which the baseline will be evaluated will be built AFTER the creation
    of the baseline. The parameters of the real environment on which the baseline will be assessed will belong to the
    same class than the argument used by the baseline. This means that if a baseline is built with a grid2op
    environment "env", this environment will not be modified in any manner, all it's internal variable will not
    change etc. This is done to prevent cheating.

    """
    def __init__(self,
                 action_space,
                 observation_space,
                 name,
                 save_path, 
                 learning_rate=1e-4,
                 gamma=0.99,
                 replace_target_iter=300,
                 replay_memory_size=300,
                 PER_alpha=0.6,
                 PER_beta=0.4,
                 batch_size = 32,
                 epsilon_start=0.7,
                 epsilon_end=0.001,
                 verbose_per_episode=1,
                 seed=22,
                 verbose=False,
                 data_dir=".",
                 model_name="geirina",
                 restore=False
                 ):
        DoNothingAgent.__init__(self, action_space)
        self.do_nothing = self.action_space()
        self.name = name
        self.agent_action_space = action_space
        self.save_path = os.path.abspath(save_path)
        if not os.path.exists(self.save_path):
            os.mkdir(self.save_path)

        self.model_name = model_name
        self.restore = restore

        self.id = set()  
        self.line_count = [0 for i in range(20)]      

        # training params
        self.n_features = np.sum(observation_space.shape)
        self.learning_rate = learning_rate
        self.gamma = gamma
        self.replace_target_iter = replace_target_iter
        self.replay_memory_size = replay_memory_size
        self.PER_alpha = PER_alpha
        self.PER_beta = PER_beta
        self.batch_size = batch_size
        self.verbose_per_episode = verbose_per_episode
        self.verbose = verbose
        # state control
        self.seed = seed
       # np.random.seed(seed)
        tf.set_random_seed(seed)

        # control epsilon
        self.epsilon_start = epsilon_start
        self.epsilon_end = epsilon_end
        

        # Directory
        # if you provide that, make sure to clearly indicate it
        # and to throw an error when it's not compatible with the environment used !
        self.data_dir = os.path.abspath(data_dir)
        action_to_index_dir = os.path.join(self.data_dir, 'action_to_index_jd1.npy')
        action_dir = os.path.join(self.data_dir, 'all_actions_jd1.npy')       

        # load actions
        self.action_space = np.load(action_dir, allow_pickle=True)
        #print("self action space", self.action_space)
        #print("self action space shape", self.action_space.shape)
        self.n_actions = self.action_space.shape[0]
        #print("self n_actions", self.n_actions)
        self.action_to_index = np.load(action_to_index_dir, allow_pickle=True)
        #print("self action to index", self.action_to_index)

        # build graph
        self.graph = tf.Graph()

        with self.graph.as_default():
          self.dqn_main = DeepQNetworkDueling(learning_rate=learning_rate,
                                              n_actions=self.n_actions,
                                              n_state=self.n_features,
                                                scope='dqn_main')
          self.dqn_target = DeepQNetworkDueling(learning_rate=learning_rate,
                                                n_actions=self.n_actions,
                                              n_state=self.n_features,
                                                scope='dqn_target')
          # Saver
          self.saver = tf.train.Saver()
          self.sess = tf.Session(graph=self.graph)
          self.sess.run(tf.global_variables_initializer())

        # The replay memory
        self.replay_memory = Memory(self.replay_memory_size, PER_alpha, PER_beta)

        # hard copy
        self.params_copy_hard = [target.assign(main) for main, target in zip(
            self.dqn_main.params_train, self.dqn_target.params_train)]

        # summary
        self.timestamp = datetime.datetime.now().strftime('%m-%d-%H-%M')
        self.episode_score_history = [0] 


    def action_transformation(self, action_array, name):
        # Transfer the l2rpn action array to a dict of grid2op
        action_dict = {}
        obj_id_dict = {}

        if name == 'l2rpn_case14_sandbox':
          self.line_map = [0,1,2,3,4,5,6,15,16,17,9,8,7,18,19,11,10,12,13,14]
          load_map = [0,1,2,3,4,5,6,7,8,9,10]
        elif name == "l2rpn_2019":
          self.line_map = [i for i in range(20)]
          load_map = [0,1,2,3,4,5,6,7,8,9,10]          
        else:
          self.line_map = [0,1,7,8,9,10,11,15,16,17,14,13,12,18,19,3,2,4,5,6]
          load_map = [0,1,3,4,5,6,7,8,9,10,2]


        # generator
        offset = 0
        switches_gen_array = action_array[:5]
        switch_gen_id_list = list(np.where(switches_gen_array == 1)[0])
        if switch_gen_id_list:
          obj_id_dict["generators_id"] = switch_gen_id_list

        # load
        offset += 5
        switches_load_array = action_array[offset:offset+11]
        switch_load_id_list = list(np.where(switches_load_array == 1)[0])
        if switch_load_id_list:
          obj_id_dict["loads_id"] = [load_map[int(i)] for i in switch_load_id_list]
          #obj_id_dict["loads_id"] = [int(i) for i in switch_load_id_list]

        # line ox
        offset += 11
        switches_lines_or_array = action_array[offset:offset+20]
        switch_or_id_list = list(np.where(switches_lines_or_array == 1)[0])

          #print(obj_id_dict["lines_or_id"])
        # line ex
        offset += 20
        switches_lines_ex_array = action_array[offset:offset+20]
        switch_ex_id_list = list(np.where(switches_lines_ex_array == 1)[0])
        
        if name == 'l2rpn_case14_sandbox':
          if 14 in switch_or_id_list:
            switch_or_id_list.remove(14)
            switch_ex_id_list.append(14) 
          elif 14 in switch_ex_id_list:
            switch_ex_id_list.remove(14)
            switch_or_id_list.append(14) 

        if switch_or_id_list:
          obj_id_dict["lines_or_id"] = [self.line_map[int(i)] for i in switch_or_id_list]
          #obj_id_dict["lines_or_id"] = [int(i) for i in switch_or_id_list]

        if switch_ex_id_list:
          obj_id_dict["lines_ex_id"] = [self.line_map[int(i)] for i in switch_ex_id_list]
        #  obj_id_dict["lines_ex_id"] = [int(i) for i in switch_ex_id_list]

        if len(obj_id_dict) != 0:
          action_dict["change_bus"] = obj_id_dict

        # line status
        switches_lines_status_array = action_array[-20:]
        switch_lines_id_list = list(np.where(switches_lines_status_array == 1)[0])
        if len(switch_lines_id_list) != 0:
          action_dict["change_line_status"] = switch_lines_id_list
        topo_action = self.agent_action_space({})
        if len(action_dict) != 0:
          try:
            topo_action = self.agent_action_space(action_dict)
          except AmbiguousAction:
            print("This Is AmbiguousAction: You can only change line status with int or boolean numpy array vector.")
        return  topo_action

    def normalize(self, state):
    	#normalize the state
        for i, x in enumerate(state[0]):
            if x >= 1e5: state[0][i] /= 1e4
            elif x >= 1e4 and x < 1e5 : state[0][i] /= 1e3
            elif x >= 1e3 and x < 1e4 : state[0][i] /= 1e2

        return(state)

    def train(self, env,
              iterations,
              logdir = "logs-train", 
              isco=''):

        self.env = env
#        self.agent_action_space = self.env.action_space
        time_start = time.time()
        self.epsilon_decay_steps = 3 * iterations // 150
        epsilons = np.append(np.linspace(self.epsilon_start, 0.2, int(self.epsilon_decay_steps/5)), np.linspace(0.19, self.epsilon_end, int(self.epsilon_decay_steps*4/5)))
        print(len(epsilons))
        count_action = {}
        

        # session config
        config = tf.ConfigProto(log_device_placement=False,
                                allow_soft_placement=True)
        # store result
        result_columns = ["Episode", "Totle Timestep", "Total Score"] 
        with open("result_summary.csv",'w',newline='') as f:
            writer = csv.writer(f)
            writer.writerow(result_columns)

        # initialize sees
        with tf.Session(config=config, graph=self.graph) as sess:
          sess.run(tf.global_variables_initializer())
          if self.restore:
            self.saver.restore(sess, (self.save_path + '/{}.ckpt'.format(self.model_name)))

          time_step = 0
          loss = 0
          episode = 0

          print("Start training...")
          sys.stdout.flush()

          done = False
          flag_next_chronic = False
          

          # start training for data in chronics
          while time_step < iterations:
            self.id = set()
            self.id.clear()

            state = self.env.reset().to_vect()
            #initialize observation
            self.observation = self.env.helper_observation(self.env)
            obs = None
            state = [0 if i != i else i for i in state]
            state = np.array(state).reshape((-1, self.n_features))
            line_count = [0 for i in range(20)]
            lines_status_list_bool = []

            for step in itertools.count():
              reward_tot = 0
              print("Epsilons is ", epsilons[min(time_step, self.epsilon_decay_steps - 1)], epsilons[self.epsilon_decay_steps - 1] )
              
              # update the target estimator
              if time_step % self.replace_target_iter == 0:
                sess.run([self.params_copy_hard])

              self.action_next = np.zeros(76)
              if step >=1:             
                  #count the cooldown time for broken line
                  for i, v in enumerate(lines_status_list_bool):
                    if v <= 0:
                      self.id.add(i)
                      line_count[i] +=1
                      if line_count[i] >10: line_count[i] = 10
                    if v > 0:
                        try: 
                            line_count[i] = 0
                            self.id.remove(i)
                        except: pass
                      
                  idlist = [i for i in self.id]
                  action_transf_dict = {}

                  #if cooldown time is reached, close the line and check validation
                  if len(idlist) > 0:   
                    print("Broken_Line", line_count)               
                    for i, x in enumerate(line_count):
                        if x == 10 and (False in lines_status_list_bool): 
                            set_status = self.agent_action_space.get_set_line_status_vect()
                            for o, e in [(1,1),(1,2),(2,1),(2,2)]:
                                action_transf_dict = self.agent_action_space.reconnect_powerline(line_id=i,bus_or=o,bus_ex=e)  
                                obs_simulate, reward_simulate, done_simulate, _ = self.observation.simulate(action_transf_dict)
                                print(action_transf_dict, done_simulate)
                                if not done_simulate: break                 

                            self.action_next[56+self.line_map.index(idlist[0])] = 1.0
                            action = self.action_to_index.item().get(tuple(self.action_next)) 
                            self.id.clear()
                            line_count[i] = 0
                            break

              #if not close line or game over in simulation, call agent
              if 1.0 not in self.action_next or done_simulate:
                  # choose action
                  action, q_predictions = self.dqn_main.act(
                      sess, self.normalize(state), epsilons[min(time_step, self.epsilon_decay_steps - 1)])
                  # guide to do nothning (action = 155)
                  if np.random.uniform() > epsilons[min(time_step, self.epsilon_decay_steps - 1)]: action = 155
                  print("training action index", action)
                  # check loadflow
                  thermal_limits = np.abs(self.env.backend.get_thermal_limit())

                  # check overflow
                  action_is_legal = True
                  has_danger = False
                  has_overflow = False              
                  action_array = self.action_space[action]
                  action_transf_dict = self.action_transformation(action_array, self.env.name)

                  obs_simulate, reward_simulate, done_simulate, infor = self.observation.simulate(action_transf_dict)
                  #print("first simulate score", reward_simulate)

                  if reward_simulate is None:
                    has_danger = True
                  elif obs is not None: 
                    has_overflow = any(obs.rho>1.02)

                  # if game over, run more simulation
                  if done_simulate or reward_simulate is None or has_danger or has_overflow:
                    # if has overflow. try all actions
                    if has_overflow:
                      print('has overflow !!!!!!!!!!!!!!!!!!!!!!!!!')
                    if has_danger:
                      print('has danger !!!!!!!!!!!!!!!!!!!!!!!!!!!')

                    top_actions = np.argsort(q_predictions)[-1: -10: -1].tolist()
                    chosen_action = 155
                    max_score = float('-inf')
                    for action in tuple(top_actions):           
                      action_class_helper = self.env.helper_action_env
                      action_array = self.action_space[action]
                      action_transf_dict = self.action_transformation(action_array, self.env.name)
                      action_is_legal = action_class_helper.legal_action(action_transf_dict, self.env)
                      if not action_is_legal:
                        #print("illegal!!!")
                        continue
                      else:
                        obs_simulate, reward_simulate, done_simulate, _= self.observation.simulate(action_transf_dict)
                        #print("simulate score", reward_simulate, action)
                      if obs_simulate is None:
                        continue
                      else:
                        obs_dict = obs_simulate.to_dict()
                        lineflow_simulate_ratio = obs_dict["rho"]

                        lineflow_simulate_ratio = [round(x, 4) for x in lineflow_simulate_ratio]

                        # seperate big line and small line
                        has_danger = False
                        for ratio, limit in zip(lineflow_simulate_ratio, thermal_limits):
                          if ratio > 1.05:
                            has_danger = True

                        if not done_simulate and reward_simulate > max_score and not has_danger:
                          max_score = reward_simulate
                          chosen_action = action
                          print('current best action: {}, score: {:.4f}'.format(chosen_action, reward_simulate))

                    # chosen action
                    action = chosen_action
              
                  # count action
                  count_action[action] = count_action.get(action, 0) + 1

                  # take a step
                  action_array = self.action_space[action]
                  action_transf_dict = self.action_transformation(action_array, self.env.name)
              #print("applied action is ", action_transf_dict)              
              obs, reward, done, infos = self.env.step(action_transf_dict)
              self.observation = obs
              #self.observation.update(self.env)

             # record data
              next_state_dict = obs.to_dict()
              lines_status_list_bool = next_state_dict["rho"]

              state_list = [0 if i != i else i for i in self.env.get_obs().to_vect()] 
              if done:
                next_state = state
                reward = -15
              else:
                next_state = np.array(state_list).reshape((-1, self.n_features))

              if reward >= 100: reward_tot = reward / 750
              else: reward_tot = reward / 15

              # record
              self.episode_score_history[episode] += reward

              # Save transition to replay memory
              if done:
                # if done: store more
                for i in range(4):
                  self.replay_memory.store(
                      [self.normalize(state), action, reward_tot, self.normalize(next_state), done])
              else:
                self.replay_memory.store(
                      [self.normalize(state), action, reward_tot, self.normalize(next_state), done])
                

              # learn
              if time_step > self.replay_memory_size and time_step % 5 == 0:
                
              # Sample a minibatch from the replay memory
                tree_idx, batch_samples, IS_weights = self.replay_memory.sample(self.batch_size)
                states_batch, action_batch, reward_batch, next_states_batch, done_batch = map(
                    np.array, zip(*batch_samples))
                states_batch = states_batch.reshape((-1, self.n_features))
                next_states_batch = next_states_batch.reshape(
                    (-1, self.n_features))

                # Calculate q targets
                q_values_next = self.dqn_target.predict(sess, next_states_batch)
                q_values_next = np.array(q_values_next[0])
                targets_batch = reward_batch + \
                    np.invert(done_batch).astype(np.float32) * \
                    self.gamma * np.amax(q_values_next, axis=1)

                # Perform gradient descent update
                loss, abs_TD_errors = self.dqn_main.update(sess, states_batch, targets_batch,
                                                           action_batch.reshape((-1, 1)), IS_weights)
                # Update priority
                self.replay_memory.batch_update(tree_idx, abs_TD_errors)

                # verbose step summary
              if episode % self.verbose_per_episode == 0 and (time_step + 1) % 1 == 0:
                print("episode: {}, step: {},  action: {}, loss: {:4f},  reward: {:4f}, score: {:.4f}, tot: {:.4f}\n".
                      format(episode + 1, step + 1, action, loss, reward, self.episode_score_history[episode], reward_tot))
                sys.stdout.flush()

              # update state
              state = next_state
              time_step += 1

              #if done or time_step > total_train_step or flag_next_chronic:
              if done: 
                break

            episode += 1
            self.episode_score_history.append(0)

            # save model per episode
            if episode >= 50:
              self.saver.save(sess, os.path.join(self.save_path, '{}.ckpt'.format(self.model_name)))
              print('Model Saved!')

            # verbose episode summary
            print("\nepisode: {}, mean_score: {:4f}, sum_score: {:4f}\n".
                      format(episode + 1, self.episode_score_history[episode] / (step + 1), self.episode_score_history[episode]))
            print("action count: {}\n".format(sorted(count_action.items())))
            with open("result_summary.csv", 'a',newline='') as f:
                result = [episode + 1]
                result.append(step+1)
                result.append(self.episode_score_history[episode])
                writer = csv.writer(f)
                writer.writerow(result)
                
          time_end = time.time()
          print("\nFinished, Total time used: {}s".format(time_end - time_start))

    def load(self, model_name):
      self.saver.restore(self.sess, os.path.join(self.save_path, '{}.ckpt'.format(model_name)))
      print('Model {} Loaded!'.format(model_name))

    def act(self, observation, reward, done=False):

      #check line:
      state_dict = observation.to_dict()
      lines_status_list_bool = state_dict["line_status"]  

      self.action_next = np.zeros(76)
      for i, v in enumerate(state_dict['rho']):
        if v <= 0:
          self.id.add(i)
          self.line_count[i] +=1
          if self.line_count[i] >10: self.line_count[i] = 10
        if v > 0:
            try: 
                self.line_count[i] = 0
                self.id.remove(i)
            except: pass
          
      idlist = [i for i in self.id]
      action_transf_dict = {}
      done_simulate = False

      #if cooldown time is reached, close the line and check validation
      if len(idlist) > 0:   
        print("Broken_Line", self.line_count)               
        for i, x in enumerate(self.line_count):
            if x == 10 and (False in lines_status_list_bool): 
                set_status = self.agent_action_space.get_set_line_status_vect()
                for o, e in [(1,1),(1,2),(2,1),(2,2)]:
                    action_transf_dict = self.agent_action_space.reconnect_powerline(line_id=i,bus_or=o,bus_ex=e)  
                    obs_simulate, reward_simulate, done_simulate, _ = observation.simulate(action_transf_dict)
                    print(action_transf_dict, done_simulate)
                    if not done_simulate: break                 

                self.action_next[56+idlist[0]] = 1.0
                action = self.action_to_index.item().get(tuple(self.action_next)) 
                self.id.clear()
                self.line_count[i] = 0
                break
      #Else agent pick action              
      if 1.0 not in self.action_next or done_simulate:

        action_array = self.action_space[155]
        action_transf_dict = self.action_transformation(action_array, self.env.name)

        obs_simulate, reward_simulate, done_simulate, infor = observation.simulate(action_transf_dict)
        #print("first simulate score", reward_simulate)
        has_danger = False
        has_overflow = False 
        if reward_simulate is None or reward_simulate <= 0:
          has_danger = True
        elif obs_simulate is not None: 
          has_overflow = any(obs_simulate.rho>1.05)

        # if game over, run more simulation
        if done_simulate or reward_simulate is None or has_danger or has_overflow:

          state = [0 if i != i else i for i in observation.to_vect()] 
          action, q_predictions = self.dqn_main.act(self.sess, self.normalize([state]), 0.0)
          top_actions = np.argsort(q_predictions)[-1: -11: -1].tolist()

          chosen_action = 155
          max_score = float('-inf')
          for action in tuple(top_actions+[155]):           

            action_array = self.action_space[action]
            action_transf_dict = self.action_transformation(action_array, self.env.name)
            obs_simulate, reward_simulate, done_simulate, _= observation.simulate(action_transf_dict)
            #print("simulate score", reward_simulate, action)
            if obs_simulate is None:
              continue
            else:
              obs_dict = obs_simulate.to_dict()
              lineflow_simulate_ratio = obs_dict["rho"]
              lineflow_simulate_ratio = [round(x, 4) for x in lineflow_simulate_ratio]

              # seperate big line and small line
              has_danger = False
              for ratio in lineflow_simulate_ratio:
                if  ratio > 1.02:
                  has_danger = True

              if not done_simulate and reward_simulate > max_score and not has_danger:
                max_score = reward_simulate
                chosen_action = action
                print('current best action: {}, score: {:.4f}'.format(chosen_action, reward_simulate))

          # chosen action
          action_array = self.action_space[chosen_action]
          action_transf_dict = self.action_transformation(action_array, self.env.name)
      #print(reward,done,action_transf_dict)
      return action_transf_dict




# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np
import random
import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.backend as K
import tensorflow.keras.models as tfkm
import tensorflow.keras.optimizers as tfko
import tensorflow.keras.layers as tfkl
import tensorflow.keras.activations as tfka


class DoubleDuelingDQN_NN(object):
    """Constructs the desired deep q learning network"""
    def __init__(self,
                 action_size,
                 observation_size,                 
                 num_frames = 4,
                 learning_rate = 1e-5,
                 learning_rate_decay_steps = 1000,
                 learning_rate_decay_rate = 0.95):
        self.action_size = action_size
        self.observation_size = observation_size
        self.lr = learning_rate
        self.lr_decay_steps = learning_rate_decay_steps
        self.lr_decay_rate = learning_rate_decay_rate
        self.num_frames = num_frames
        self.model = None
        self.construct_q_network()

    def construct_q_network(self):
        input_shape = (self.observation_size * self.num_frames,)
        input_layer = tfk.Input(shape = input_shape, name="input_obs")
        lay1 = tfkl.Dense(self.observation_size * 2, name="fc_1")(input_layer)
        lay1 = tfka.relu(lay1, alpha=0.01) #leaky_relu

        lay2 = tfkl.Dense(self.observation_size, name="fc_2")(lay1)
        lay2 = tfka.relu(lay2, alpha=0.01) #leaky_relu

        lay3 = tfkl.Dense(896, name="fc_3")(lay2)
        lay3 = tfka.relu(lay3, alpha=0.01) #leaky_relu

        lay4 = tfkl.Dense(512, name="fc_4")(lay3)
        lay4 = tfka.relu(lay4, alpha=0.01) #leaky_relu

        advantage = tfkl.Dense(384, name="fc_adv")(lay4)
        advantage = tfka.relu(advantage, alpha=0.01) #leaky_relu
        advantage = tfkl.Dense(self.action_size, name="adv")(advantage)
        advantage_mean = tf.math.reduce_mean(advantage,
                                             axis=1, keepdims=True,
                                             name="adv_mean")
        advantage = tfkl.subtract([advantage, advantage_mean],
                                  name="adv_subtract")

        value = tfkl.Dense(384, name="fc_val")(lay4)
        value = tfka.relu(value, alpha=0.01) #leaky_relu
        value = tfkl.Dense(1, name="val")(value)

        Q = tf.math.add(value, advantage, name="Qout")

        self.model = tfk.Model(inputs=[input_layer], outputs=[Q],
                               name=self.__class__.__name__)

        # Backwards pass
        self.schedule = tfko.schedules.InverseTimeDecay(self.lr,
                                                        self.lr_decay_steps,
                                                        self.lr_decay_rate)
        self.optimizer = tfko.Adam(learning_rate=self.schedule, clipnorm=1.0)

    def train_on_batch(self, x, y_true, sample_weight):
        with tf.GradientTape() as tape:
            # Get y_pred for batch
            y_pred = self.model(x)

            # Compute loss for each sample in the batch
            batch_loss = self._batch_loss(y_true, y_pred)
        
            # Apply samples weights
            tf_sample_weight = tf.convert_to_tensor(sample_weight,
                                                    dtype=tf.float32)
            batch_loss = tf.math.multiply(batch_loss, tf_sample_weight)
            
            # Compute mean scalar loss
            loss = tf.math.reduce_mean(batch_loss)

        # Compute gradients
        grads = tape.gradient(loss, self.model.trainable_variables)

        # Apply gradients
        grad_pairs = zip(grads, self.model.trainable_variables)
        self.optimizer.apply_gradients(grad_pairs)

        # Store LR
        self.train_lr = self.optimizer._decayed_lr('float32').numpy()
        # Return loss scalar
        return loss.numpy()

    def _batch_loss(self, y_true, y_pred):
        sq_error = tf.math.square(y_true - y_pred, name="sq_error")

        # We store it because that's the priorities vector
        # for importance update
        batch_sq_error = tf.math.reduce_sum(sq_error, axis=1,
                                            name="batch_sq_error")
        # Stored as numpy array since we are in eager mode
        self.batch_sq_error = batch_sq_error.numpy()

        return batch_sq_error

    def random_move(self):
        opt_policy = np.random.randint(0, self.action_size)

        return opt_policy
        
    def predict_move(self, data):
        model_input = data.reshape(1, self.observation_size * self.num_frames)
        q_actions = self.model.predict(model_input, batch_size = 1)     
        opt_policy = np.argmax(q_actions)

        return opt_policy, q_actions[0]

    def update_target_hard(self, target_model):
        this_weights = self.model.get_weights()
        target_model.set_weights(this_weights)

    def update_target_soft(self, target_model, tau=1e-2):
        tau_inv = 1.0 - tau
        # Get parameters to update
        target_params = target_model.trainable_variables
        main_params = self.model.trainable_variables

        # Update each param
        for i, var in enumerate(target_params):
            var_persist = var.value() * tau_inv
            var_update = main_params[i].value() * tau
            # Poliak averaging
            var.assign(var_update + var_persist)

    def save_network(self, path):
        # Saves model at specified path as h5 file
        self.model.save(path)
        print("Successfully saved model at: {}".format(path))

    def load_network(self, path):
        # Load from a model.h5 file
        self.model.load_weights(path)
        print("Successfully loaded network from: {}".format(path))




#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import argparse
import tensorflow as tf

from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQN import DoubleDuelingDQN as D3QNAgent
from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQNConfig import DoubleDuelingDQNConfig as D3QNConfig

DEFAULT_NAME = "DoubleDuelingDQN"
DEFAULT_SAVE_DIR = "./models"
DEFAULT_LOG_DIR = "./logs-train"
DEFAULT_PRE_STEPS = 256
DEFAULT_TRAIN_STEPS = 1024
DEFAULT_N_FRAMES = 4
DEFAULT_BATCH_SIZE = 32
DEFAULT_LR = 1e-5
DEFAULT_VERBOSE = True

def cli():
    parser = argparse.ArgumentParser(description="Train baseline DDQN")
    # Paths
    parser.add_argument("--name", default=DEFAULT_NAME,
                        help="The name of the model")
    parser.add_argument("--data_dir", default="rte_case14_realistic",
                        help="Path to the dataset root directory")
    parser.add_argument("--save_dir", required=False,
                        default=DEFAULT_SAVE_DIR, type=str,
                        help="Directory where to save the model")
    parser.add_argument("--load_file", required=False,
                        help="Path to model.h5 to resume training with")
    parser.add_argument("--logs_dir", required=False,
                        default=DEFAULT_LOG_DIR, type=str,
                        help="Directory to save the logs")
    # Params
    parser.add_argument("--num_pre_steps", required=False,
                        default=DEFAULT_PRE_STEPS, type=int,
                        help="Number of random steps before training")
    parser.add_argument("--num_train_steps", required=False,
                        default=DEFAULT_TRAIN_STEPS, type=int,
                        help="Number of training iterations")
    parser.add_argument("--num_frames", required=False,
                        default=DEFAULT_N_FRAMES, type=int,
                        help="Number of stacked states to use during training")
    parser.add_argument("--batch_size", required=False,
                        default=DEFAULT_BATCH_SIZE, type=int,
                        help="Mini batch size (defaults to 1)")
    parser.add_argument("--learning_rate", required=False,
                        default=DEFAULT_LR, type=float,
                        help="Learning rate for the Adam optimizer")
    return parser.parse_args()


def train(env,
          name = DEFAULT_NAME,
          iterations = DEFAULT_TRAIN_STEPS,
          save_path = DEFAULT_SAVE_DIR,
          load_path = None,
          logs_path = DEFAULT_LOG_DIR,
          num_pre_training_steps = DEFAULT_PRE_STEPS,
          num_frames = DEFAULT_N_FRAMES,
          batch_size= DEFAULT_BATCH_SIZE,
          learning_rate= DEFAULT_LR,
          verbose=DEFAULT_VERBOSE):

    # Set config
    D3QNConfig.LR = learning_rate
    D3QNConfig.N_FRAMES = num_frames
    D3QNConfig.BATCH_SIZE = batch_size
    D3QNConfig.VERBOSE = verbose

    # Limit gpu usage
    physical_devices = tf.config.list_physical_devices('GPU')
    if len(physical_devices) > 0:
        tf.config.experimental.set_memory_growth(physical_devices[0], True)

    agent = D3QNAgent(env.observation_space,
                      env.action_space,
                      name=name,
                      is_training=True)

    if load_path is not None:
        agent.load(load_path)

    agent.train(env,
                iterations,
                save_path,
                num_pre_training_steps,
                logs_path)


if __name__ == "__main__":
    from grid2op.MakeEnv import make
    from grid2op.Reward import *
    from grid2op.Action import *
    from grid2op.Parameters import Parameters
    import sys

    args = cli()

    # Create grid2op game environement
    env = make(args.data_dir,
               action_class=TopologyChangeAndDispatchAction,
               reward_class=CombinedScaledReward)

    # Only load 128 steps in ram
    env.chronics_handler.set_chunk_size(128)

    # Register custom reward for training
    cr = env.reward_helper.template_reward
    #cr.addReward("overflow", CloseToOverflowReward(), 1.0)
    cr.addReward("game", GameplayReward(), 1.0)
    #cr.addReward("recolines", LinesReconnectedReward(), 1.0)
    cr.addReward("l2rpn", L2RPNReward(), 2.0/float(env.n_line))
    # Initialize custom rewards
    cr.initialize(env)
    # Set reward range to something managable
    cr.set_range(-1.0, 1.0)

    train(env,
          name = args.name,
          iterations = args.num_train_steps,
          num_pre_training_steps = args.num_pre_steps,
          save_path = args.save_dir,
          load_path = args.load_file,
          logs_path = args.logs_dir,
          num_frames = args.num_frames,
          batch_size = args.batch_size,
          learning_rate = args.learning_rate)



import logging
from copy import deepcopy

import numpy as np

import grid2op
import l2rpn_baselines.PandapowerOPFAgent.pp_functions as ppf
import pandapower as pp
from grid2op.Agent import BaseAgent


def make_logger():
    with open("agent_logs.txt", "w") as f:
        f.write("")
    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.FileHandler("result.txt"))
    _logger.addHandler(logging.StreamHandler())
    _logger.setLevel(logging.INFO)
    return _logger


class PandapowerOPFAgent(BaseAgent):
    """
    The purpose of this agent is to supply a usable base for power system calculations based on the grid2op backend.
    Incoming observations are automatically parsed into a pandapower grid. The grid can then be used for any
    algorithms based on it, e.g. optimal power flow.
    We show
    """
    timestep = 0
    last_opf_ts = 0
    grid = None
    baseline_losses = None
    total_losses = 0.
    gen_max_p = None
    gen_min_p = None
    logger = make_logger()

    def __init__(self, action_space, grid_path: str,
                 acceptable_loading_pct: float = 99.9,
                 min_loss_reduction_mwt: float = 0.5,
                 opf_type: str = "pypower",
                 line_auto_reconnect: bool = False):
        """
        Initialize agent
        :param action_space: the Grid2Op action space
        :param grid_path: path to the pandapower grid as a json file
        :param acceptable_loading_pct: maximum line/transformer loading allowed before OPF engages
        :param min_loss_reduction_mwt: minimum loss (as MW * timestep unit t) that needs to be reduced for the OPF to engage
        :param opf_type: which OPF to use (pypower or powermodels)
        :param line_auto_reconnect: reconnect disconnected lines automatically as soon as possible
        """
        BaseAgent.__init__(self, action_space)
        self.do_nothing_action = action_space({})
        self.grid = pp.from_json(grid_path)
        self.acceptable_loading_pct = acceptable_loading_pct
        self.min_loss_reduction_mwt = min_loss_reduction_mwt
        self.opf_type = opf_type
        self.line_auto_reconnect = line_auto_reconnect
        assert self.opf_type.lower() in ("pypower", "powermodels"), "choose either pypower or powermodels as the opf type"

    def parse_observation_to_grid(self, obs: grid2op.Observation):
        if self.timestep == 0:
            self.grid.trafo["max_loading_percent"] = self.acceptable_loading_pct
            self.grid.line["max_loading_percent"] = self.acceptable_loading_pct
            self.grid.poly_cost.drop(self.grid.poly_cost.index, inplace=True)
            assert len(self.grid.ext_grid) == 1 and len(obs.gen_type) == len(self.grid.gen) + len(self.grid.ext_grid)
            self.grid.bus.min_vm_pu = 0.9
            self.grid.bus.max_vm_pu = 1.2
            # gen
            self.grid.gen.type = obs.gen_type[:-1]
            self.grid.gen.controllable = obs.gen_redispatchable[:-1]
            self.grid.gen.min_p_mw = obs.gen_pmin[:-1]
            self.grid.gen.max_p_mw = obs.gen_pmax[:-1]
            self.gen_min_p = obs.gen_pmin
            self.gen_max_p = obs.gen_pmax
            # ext grid (last gen)
            self.grid.gen.min_q_mvar = -obs.gen_pmax[:-1]
            self.grid.gen.max_q_mvar = obs.gen_pmax[:-1]
            self.grid.ext_grid.type = obs.gen_type[-1]
            assert obs.gen_redispatchable[-1]
            self.grid.ext_grid.min_p_mw = obs.gen_pmin[-1]
            self.grid.ext_grid.max_p_mw = obs.gen_pmax[-1]
            self.grid.ext_grid.min_q_mvar = -obs.gen_pmax[-1]
            self.grid.ext_grid.max_q_mvar = obs.gen_pmax[-1]
            for row in self.grid.gen[self.grid.gen.controllable].itertuples():
                pp.create_poly_cost(self.grid, row.Index, "gen", cp1_eur_per_mw=1)  # perform loss minimization
            pp.create_poly_cost(self.grid, self.grid.ext_grid.index[0], "ext_grid", cp1_eur_per_mw=1)
        self.refresh_gen_values(obs)
        self.grid.load["p_mw"] = obs.load_p
        self.grid.load["q_mvar"] = obs.load_q
        self.grid.line["in_service"] = obs.line_status[:len(self.grid.line)]
        self.grid.trafo["in_service"] = obs.line_status[-len(self.grid.trafo):]
        pp.runpp(self.grid, init="results")
        self.timestep += 1
        self.baseline_losses = ppf.calc_losses(self.grid)
        self.total_losses += self.baseline_losses
        if (self.grid.res_line.loading_percent > self.acceptable_loading_pct).any() \
                or (self.grid.res_trafo.loading_percent > self.acceptable_loading_pct).any():
            self.logger.info(f"OVERLOADING: line @ {self.grid.res_line.loading_percent.max():.1f} % "
                             f"/ trafo @ {(self.grid.res_trafo.loading_percent / self.grid.trafo.max_loading_percent).max() * 100.:.1f} %")

    def refresh_gen_values(self, obs: grid2op.Observation):
        # update pandapower gen values with observation
        for row in self.grid.ext_grid.itertuples():
            self.grid.ext_grid.loc[row.Index, "vm_pu"] = ppf.find_value_for_pp_bus(self.grid, obs.v_or, obs.v_ex, row.bus)
        for row in self.grid.gen.itertuples():
            self.grid.gen.loc[row.Index, "vm_pu"] = ppf.find_value_for_pp_bus(self.grid, obs.v_or, obs.v_ex, row.bus)
        self.grid.gen["p_mw"] = obs.prod_p[:-1]
        self.grid.gen["res_q_mvar"] = obs.prod_q[:-1]
        ramp_down, ramp_up = obs.gen_max_ramp_down + 1e-6, obs.gen_max_ramp_up - 1e-6
        abs_min, abs_max = np.minimum(self.grid.gen.p_mw, obs.gen_pmin[:-1]), np.maximum(self.grid.gen.p_mw, obs.gen_pmax[:-1])
        self.grid.gen["max_p_mw"] = (self.grid.gen.p_mw + ramp_up[:-1]).clip(abs_min, abs_max)
        self.grid.gen["min_p_mw"] = (self.grid.gen.p_mw - ramp_down[:-1]).clip(abs_min, abs_max)
        self.grid.ext_grid["p_mw"] = obs.prod_p[-1]
        self.grid.ext_grid["max_p_mw"] = (obs.prod_p[-1] + ramp_up[-1]).clip(obs.gen_pmin[-1], obs.gen_pmax[-1])
        self.grid.ext_grid["min_p_mw"] = (obs.prod_p[-1] + ramp_up[-1]).clip(obs.gen_pmin[-1], obs.gen_pmax[-1])

    def act(self, observation: grid2op.Observation, reward, done=False):
        # 1. Parse observations into pandapower grid
        self.parse_observation_to_grid(observation)
        # 2. Check if any lines are out of service in the observed grid and mark them (so they are not changed with an action)
        #    Lines that were out of service but can be reconnected will automatically be connected again with an action
        #    Grid2Op lines can both be pandapower lines or pandapower transformers
        opf_grid = deepcopy(self.grid)
        lines_to_be_connected, lines_to_disconnect, line_failures = [], [], []
        if not self.grid.line.in_service.all():
            for line_oos in self.grid.line.loc[~self.grid.line.in_service].index:
                line_zero_idx = self.grid.line.index.get_loc(line_oos)
                line_state = observation.state_of(line_id=line_zero_idx)
                if line_state["indisponibility"] > 0:
                    opf_grid.line.drop(line_oos, inplace=True)  # drop line, otherwise it would be used in the OPF
                    self.logger.info(f"Observation: Line {line_oos} out of service!")
                    line_failures.append(line_zero_idx)
                elif self.line_auto_reconnect:
                    lines_to_be_connected.append(line_zero_idx)  # automatically reconnect lines if possible
                    self.grid.line.loc[line_oos, "in_service"] = True
        if not self.grid.trafo.in_service.all():
            for trafo_oos in self.grid.trafo.loc[~self.grid.trafo.in_service].index:
                trafo_zero_idx = self.grid.trafo.index.get_loc(trafo_oos) + len(self.grid.line)  # offset pp lines
                trafo_state = observation.state_of(line_id=trafo_zero_idx)
                if trafo_state["indisponibility"] > 0:
                    opf_grid.trafo.drop(trafo_oos, inplace=True)  # drop trafo, otherwise it would be used in the OPF
                    self.logger.info(f"Observation: trafo {trafo_oos} out of service!")
                    line_failures.append(trafo_zero_idx)
                elif self.line_auto_reconnect:
                    lines_to_be_connected.append(trafo_zero_idx)  # automatically reconnect trafos if possible
                    self.grid.trafo.loc[trafo_oos, "in_service"] = True
        # 3. Perform OPF
        p_dispatched_current = np.array(self.grid.gen.p_mw.tolist() + self.grid.ext_grid.p_mw.tolist())
        p_redispatched = p_dispatched_current
        generation, opf_line_status, opf_trafo_status, make_changes = ppf.run_opf(opf_grid, self.min_loss_reduction_mwt, self.acceptable_loading_pct,
                                                                                  self.opf_type.lower(), logger=self.logger)
        if make_changes:
            lines_to_disconnect = ppf.make_zero_idx(self.grid.line, opf_line_status.index[~opf_line_status])
            lines_to_disconnect.extend(ppf.make_zero_idx(self.grid.trafo, opf_trafo_status.index[~opf_trafo_status], offset=len(self.grid.line)))
            lines_to_be_connected = ppf.make_zero_idx(self.grid.line, set(self.grid.line.index[~self.grid.line.in_service])
                                                       & set(opf_line_status.index[opf_line_status]))
            lines_to_be_connected.extend(ppf.make_zero_idx(self.grid.trafo, set(self.grid.trafo.index[~self.grid.trafo.in_service])
                                                            & set(opf_trafo_status.index[opf_trafo_status]), offset=len(self.grid.line)))
            new_connections = sorted(list(set(self.grid.line.index[~self.grid.line.in_service]) & set(opf_line_status.index[opf_line_status])))
            self.logger.info(f"OPF wants to disconnect: pp_{opf_line_status.index[~opf_line_status].tolist()} "
                             f"/ connect: pp_{self.grid.line.index[new_connections].tolist()} ")
            p_redispatched = generation
        # 4. Convert to actions
        action_space = {}
        # Redispatch action
        gen_p_redispatched = ~np.isclose(p_dispatched_current, p_redispatched)
        if np.any(gen_p_redispatched):
            p_redispatched -= p_dispatched_current  # make absolute power relative
            p_redispatched[abs(p_redispatched) < 0.5] = 0.
            redispatch_sum = p_redispatched[gen_p_redispatched].sum()  # this sum needs to be 0. for a valid redispatch
            if redispatch_sum > 0.:
                power_up = np.where(p_redispatched > 0)[0]
                p_redispatched[power_up] -= redispatch_sum / len(power_up)
            elif redispatch_sum < 0.:
                power_down = np.where(p_redispatched < 0)[0]
                p_redispatched[power_down] -= redispatch_sum / len(power_down)
            assert abs(p_redispatched[gen_p_redispatched].sum()) < 1e-14
            action_space["redispatch"] = [(idx, p) for idx, p in zip(np.arange(len(self.grid.gen) + 1)[gen_p_redispatched],
                                                                     p_redispatched[gen_p_redispatched]) if abs(p) >= 1e-3]
            # self.logger.info(action_space["redispatch"])
        # Set lines action
        new_line_status_array = np.zeros(observation.rho.shape)
        bus_idx = {"lines_or_id": [], "lines_ex_id": []}
        assert len(set(lines_to_disconnect) & set(lines_to_be_connected)) == 0
        if len(lines_to_disconnect) > 0:
            # new_line_status_array[np.random.choice(lines_to_disconnect)] = -1  # can only open 1 line per time step (select randomly)
            new_line_status_array[lines_to_disconnect[-1]] = -1  # can only open 1 line per time step (select transformers first)
        for line_idx in lines_to_be_connected:
            line_state = observation.state_of(line_id=line_idx)
            if line_state["cooldown_time"] > 0:
                self.logger.error(f"Cannot reconnect line {line_idx}, it is still in cooldown!")
                continue
            new_line_status_array[line_idx] = 1
            bus_idx["lines_or_id"].append((line_idx, 1))
            bus_idx["lines_ex_id"].append((line_idx, 1))
        if np.count_nonzero(new_line_status_array) > 0:
            action_space["set_line_status"] = new_line_status_array
            action_space["set_bus"] = bus_idx
        action = self.action_space(action_space)
        # 5. Simulate action to see whether it makes sense, otherwise do nothing
        # if action != self.do_nothing_action:
        #     simul_obs, simul_reward, simul_has_error, simul_info = observation.simulate(action)
        #     if simul_has_error:
        #         action = self.do_nothing_action
        #     else:
        #         self.logger.info(action)
        #         _, dnr, _, _ = observation.simulate(self.do_nothing_action)
        #         self.logger.info(f"Expected reward: {simul_reward:.2f} (do nothing: {dnr:.2f})")
        # 6. Send action, done
        # return self.do_nothing_action
        denied, reason = action.is_ambiguous()
        assert not denied, f"{reason} {action}"
        return action



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

_WARN_GPU_MEMORY = "Impossible to locate tensorflow GPU devices. If you have GPUs on your systems, this script " \
                   "might take all their available memory."


# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

# tf2.0 friendly
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Activation, Dense
    from tensorflow.keras.layers import Input, Lambda, subtract, add
    import tensorflow.keras.backend as K

from l2rpn_baselines.utils import BaseDeepQ, TrainingParam


class DuelQ_NN(BaseDeepQ):
    """Constructs the desired duelling deep q learning network"""
    def __init__(self,
                 nn_params,
                 training_param=None):
        if training_param is None:
            training_param = TrainingParam()
        BaseDeepQ.__init__(self,
                           nn_params,
                           training_param)
        self.construct_q_network()

    def construct_q_network(self):
        """
        It uses the architecture defined in the `nn_archi` attributes.

        """
        self._model = Sequential()
        input_layer = Input(shape=(self._nn_archi.observation_size,),
                            name="observation")

        lay = input_layer
        for lay_num, (size, act) in enumerate(zip(self._nn_archi.sizes, self._nn_archi.activs)):
            lay = Dense(size, name="layer_{}".format(lay_num))(lay)  # put at self.action_size
            lay = Activation(act)(lay)

        fc1 = Dense(self._action_size)(lay)
        advantage = Dense(self._action_size, name="advantage")(fc1)

        fc2 = Dense(self._action_size)(lay)
        value = Dense(1, name="value")(fc2)

        meaner = Lambda(lambda x: K.mean(x, axis=1))
        mn_ = meaner(advantage)
        tmp = subtract([advantage, mn_])
        policy = add([tmp, value], name="policy")

        self._model = Model(inputs=[input_layer], outputs=[policy])
        self._schedule_model, self._optimizer_model = self.make_optimiser()
        self._model.compile(loss='mse', optimizer=self._optimizer_model)

        self._target_model = Model(inputs=[input_layer], outputs=[policy])
